import pandas as pd
import numpy as np
from pyadlml.dataset._core.acts_and_devs import label_data
from pyadlml.dataset._core.devices import create_device_info_dict
from pyadlml.dataset.stats.activities import _get_freq_func
from pyadlml.util import get_npartitions, get_parallel
from pyadlml.constants import START_TIME, END_TIME, TIME, DEVICE, VALUE, ACTIVITY, CAT, NUM, BOOL
from pyadlml.dataset.util import infer_dtypes, categorical_2_binary
import dask.dataframe as dd
from pyadlml.dataset._representations.raw import create_raw
#import __logger__


def cross_correlogram(df_devices, df_activities, maxlag, binsize, other=False):
    """
    Computes the cross-correlogram between activity beginning and ending and device events.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    df_activities : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    other : bool, default=False
        Determines whether gaps between activities should be assigned
        the activity *other* or be ignored.

    Returns
    -------
    pd.DataFrame

    """
    ET = 'event_times'
    df_dev = df_devices.copy()
    df_act = df_activities.copy()

    devices = df_devices[DEVICE].unique()
    activities = df_activities[ACTIVITY].unique()

    maxlag = pd.Timedelta(maxlag).seconds
    binsize = pd.Timedelta(binsize).seconds

    n_bins = int((maxlag / binsize) * 2) + 1    # add +1 for symmetric histogram
    n_dev = len(devices)
    n_act = len(activities)

    ccg = np.zeros((n_dev, n_act, n_bins))
    bins = np.linspace(-maxlag, maxlag, n_bins)


    df_act_st = df_act.loc[:, [START_TIME, ACTIVITY]]
    df_act_et = df_act.loc[:, [END_TIME, ACTIVITY]]

    # get times in milliseconds relative to the first event
    start_time = min(df_act_st[START_TIME].iloc[0], df_dev[TIME].iloc[0])

    df_dev[ET] = (df_dev[TIME] - start_time).dt.seconds
    df_act_st[ET] = (df_act_st[START_TIME] - start_time).dt.seconds
    df_act_et[ET] = (df_act_et[END_TIME] - start_time).dt.seconds

    def calc_hist(df_dev, df_act, device, activity):
        # select reference and target device
        t_ref = df_dev[df_dev[DEVICE] == device][ET].values
        t_tar = df_act[df_act[ACTIVITY] == activity][ET].values

        tar_matrix = np.tile(t_tar, (len(t_ref), 1))
        shift = np.repeat(t_ref.reshape(-1, 1), len(t_tar), axis=1)
        shifted_target = tar_matrix - shift

        tmp = np.apply_along_axis(np.histogram, axis=1, arr=shifted_target,
                                  bins=n_bins, range=(-maxlag, maxlag), density=False)
        hist_sum = tmp[:, 0].sum()

        return hist_sum


    for i, j in zip(range(n_dev), range(n_act)):
        result_right = calc_hist(df_dev, df_act_st, devices[i], activities[j])
        result_left = calc_hist(df_dev, df_act_et, devices[i], activities[j])

        split = int(np.floor(n_bins/2))
        ccg[i, j, :] = np.concatenate([result_right[:split], [0], result_left[split+1:]])

    return ccg, bins


def contingency_table_events(df_devices, df_activities, per_state=False, other=False, n_jobs=1):
    """
    Compute the amount of device triggers occuring during the different activities.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    df_activities : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    per_state : bool, default=False
        Determines whether events are plotted
    other : bool, default=False
        Determines whether gaps between activities should be assigned
        the activity *other* or be ignored.
    n_jobs : int, default=1
        The numberof parallel threads to start for computing the statistics

    Examples
    --------
    >>> from pyadlml.stats import contingency_events
    >>> contingency_events(data.df_devices, data.df_activities)
    activity            get drink  go to bed  ...  use toilet
    device
    Cups cupboard              36          0  ...           0
    Dishwasher                  2          0  ...           0
               ...            ...        ...  ...         ...
    Washingmachine              0          0  ...           0
    [7 rows x 7 columns]

    Results
    -------
    df : pd.DataFrame
    """
    ON = 'on'
    OFF = 'off'

    # Save original devices and activities
    devs = df_devices[DEVICE].unique()
    acts = df_activities[ACTIVITY].unique()

    df = label_data(df_devices, df_activities, other=other, n_jobs=n_jobs)
    if not per_state:
        df[VALUE] = 1
        df = pd.pivot_table(df, columns=ACTIVITY, index=DEVICE, values=VALUE, aggfunc=len, fill_value=0)\

        # Add activities where no event occurs to table
        if len(df.columns) != len(acts):
            for act in set(acts).difference(set(df.columns)):
                df[act] = 0

        # Add devices that don't hit any activity to table
        if len(df.index) != len(devs):
            for dev in set(devs).difference(set(df.index)):
                df.loc[dev] = [0]*len(acts)

        return df.reset_index()
    else:
        dtypes = infer_dtypes(df_devices)
        df = categorical_2_binary(df, dtypes[CAT])
        df.loc[df[DEVICE].isin(dtypes[NUM]), VALUE] = True

        # drop categories report of off
        mask = (df[DEVICE].isin(dtypes[NUM]) | df[DEVICE].isin(dtypes[BOOL]) | df[VALUE] == True)
        df = df.loc[mask, :]

        df['val2'] = df[VALUE].astype(int)
        df = pd.pivot_table(df, columns=ACTIVITY, index=[DEVICE, VALUE], values='val2',
                            aggfunc=len, fill_value=0)

        # rename labels to include states
        df = df.reset_index()
        df['label_postfix'] = df[VALUE].map({False: OFF, True: ON})
        bool_mask = df[DEVICE].isin(dtypes[BOOL])
        df.loc[bool_mask, DEVICE] = df.loc[bool_mask, DEVICE] + ':' + df.loc[bool_mask, 'label_postfix']

        return df.drop([VALUE, 'label_postfix'], axis=1)


def _join_to_interval(x):
    #print('x: ', x)
    return pd.Interval(x[0], x[1])


def _create_cont_df():
    """
    returns: empty pd Dataframe 
        | i | device | val |  activity | time_diff
        -----------------------------
          1 | dev1   | 0   | act1      | td
    """
    td = 'time_diff'
    df = pd.DataFrame(columns=[DEVICE, VALUE, ACTIVITY, td])
    df[td] = pd.to_timedelta(df[td])
    return df 


def _calc_intervall_diff(i1, i2):
    """
    computes the difference of time between two overlapping intervals
        form new intervall by taking 
        #  |...| 
        #  ~  ~
        # |...|
    
    returns
    -------
        time difference 
    """
    assert i1.overlaps(i2)
    mi = max([i1.left, i2.left])
    ma = min([i1.right, i2.right])
    #return pd.Interval(mi, ma)
    return ma-mi


def contingency_table_states_old(df_devs, df_acts, other=False, distributed=False):
    """
    Compute the time a device is "on" or "off" respectively
    during the different activities.

    Parameters
    ----------
    df_devs : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    other : bool
        Determines whether gaps between activities should be assigned
        the activity *other* or be ignored.

    Examples
    --------
    >>> from pyadlml.stats import contingency_duration
    >>> contingency_duration(data.df_devs, data.df_activities)
    activity                     get drink ...             use toilet
    Hall-Bedroom door Off  0 days 00:01:54 ... 0 days 00:12:24.990000
    Hall-Bedroom door On   0 days 00:14:48 ... 0 days 03:02:49.984000
    ...                                ...
    Washingmachine On      0 days 00:00:00 ...        0 days 00:00:00
    [14 rows x 7 columns]

    Returns
    -------
    df : pd.DataFrame
    """
    TD = 'td'
    ON_praefix = 'on'
    OFF_praefix = 'off'
    SEP = ':'
    
    def func(row, raw):
        """ Determines for each activity row the total time that was spent in a state for each device
        Parameters
        ----------
        row : pd.Series
            A activity dataframes row containing the start and end time for one activity
        """        
        act_start_time = row.start_time
        act_end_time = row.end_time

        # Get selection of devices that fall into the activities time frame
        raw_sel = raw[(act_start_time <= raw[TIME]) & (raw[TIME] <= act_end_time)].copy()

        if raw_sel.empty:
            # Correct for intervals when no event falls into
            #           I -- activity -- I
            #  i~~~~~~~~|~~~~~~~~~~~~~~~~|~~i
            # Get first state vector preceding the interval, since this state is valid for
            # the whole activity. Then set the time difference to the activity length
            start_row = raw[(raw[TIME] <= act_start_time)].iloc[-1]
            raw_sel = raw_sel.append(start_row)
            raw_sel.iat[0, raw_sel.columns.get_loc(TD)] = act_end_time - act_start_time
        else:

            # correct for the intervals before/after the first/last events state vector
            #               I --activity -- ... --------I
            #     i ~~~~~~~~|~~~ i ...         ... i ~~~|~~~~~ i

            # try to get the preceding state vector of devices before the activity starts
            idx_pre_sel = raw_sel.index[0] - 1
            if idx_pre_sel == -1:
                # Edge case when the first activity starts before the first recording
                # I--activity---- ...
                # i~~~tda~~~i~~ ...
                # Clone first event to activity start. Only one device has wrong state for that amount of time.
                raw_sel = raw_sel.append(raw_sel.iloc[0].copy()).sort_values(by=[TIME])
                raw_sel.iat[0, raw_sel.columns.get_loc(TD)] = raw_sel.iloc[1].time - act_start_time  # tda
            else:
                #         I--- activity ---- ....
                # i~~~~~td~~~~~~i~~~ ...
                # |--tdta-|-----|
                raw_sel = raw_sel.append(raw.iloc[idx_pre_sel]).sort_values(by=[TIME])
                raw_sel.at[raw_sel.iloc[0].name, TD] -= (act_start_time - raw_sel.iloc[0].time)  # td - tdta

            # shorten the last state vectors time difference within the activity frame to the activity's ends
            raw_sel.at[raw_sel.iloc[-1].name, TD] = act_end_time - raw_sel.iloc[-1].time


        for dev in row.index[3:]:
            state_agg = raw_sel[[dev, TD]].groupby(by=[dev])[TD].sum()
            # fails if device was not in that state for the whole activity frame
            try:
                time = state_agg.loc[True]
            except KeyError:
                time = pd.Timedelta('0ns')
            row.at[dev] = time

        return row

    dtypes = infer_dtypes(df_devs)

    # drop numerical devices
    df_devs = df_devs.loc[~df_devs[DEVICE].isin(dtypes[NUM]), :]

    dataset_info = create_device_info_dict(df_devs)
    df_devs = df_devs.sort_values(by=TIME)
    raw = create_raw(df_devs, dataset_info)

    # raw representation one-hot-encode categorical devices and create on-boolean and off-boolean devices
    raw = pd.get_dummies(data=raw, columns=dtypes[CAT], prefix_sep=SEP, dtype=bool)

    for dev in dtypes[BOOL]:
        raw[dev + SEP + OFF_praefix] = ~raw[dev]
        raw = raw.rename(columns={dev: dev + SEP + ON_praefix})
    raw[TD] = raw[TIME].shift(-1) - raw[TIME]

    df_acts = df_acts.copy().join(pd.DataFrame(index=df_acts.index, columns=raw.columns[1:-1]))

    if distributed:
        df = df_acts.parallel_apply(func, args=[raw], axis=1)
    else:
        df = df_acts.apply(func, args=[raw], axis=1)
    df = df.drop(columns=[START_TIME, END_TIME])
    df = df.groupby(ACTIVITY).sum()

    return df.T

    if True:

        import dask.dataframe as dd
        def create_meta(raw):
            devices = {name: 'object' for name in raw.columns[1:-1]}
            return {**{START_TIME: 'datetime64[ns]', END_TIME: 'timedelta64[ns]', ACTIVITY: 'object'}, **devices}
        meta = create_meta(raw)
        #   After:  .apply(func, meta={'start_time': 'float64', 'end_time': 'float64', 'activity': 'float64', 'cutlery drawer kwick:on': 'float64', 'kitchen pir:on': 'float64', 'kwik dresser:on': 'float64', 'toilet door:on': 'float64', 'microwave:on': 'float64', 'kwik stove lid:on': 'float64', 'Bedroom door:on': 'float64', 'balcony door:on': 'float64', 'sink float:on': 'float64', 'toaster:on': 'float64', 'bathroom pir:on': 'float64', 'pressure mat server corner:on': 'float64', 'pressure mat office chair:on': 'float64', 'press bed left:on': 'float64', 'toilet flush:on': 'float64', 'cupboard groceries:on': 'float64', 'cupboard plates:on': 'float64', 'fridge:on': 'float64', 'frame:on': 'float64', 'press bed right:on': 'float64', 'frontdoor:on': 'float64', 'bedroom pir:on': 'float64', 'toilet door:off': 'float64', 'kwik dresser:off': 'float64', 'frontdoor:off': 'float64', 'cutlery drawer kwick:off': 'float64', 'bedroom pir:off': 'float64', 'press bed left:off': 'float64', 'kwik stove lid:off': 'float64', 'bathroom pir:off': 'float64', 'toilet flush:off': 'float64', 'cupboard groceries:off': 'float64', 'cupboard plates:off': 'float64', 'Bedroom door:off': 'float64', 'fridge:off': 'float64', 'pressure mat office chair:off': 'float64', 'sink float:off': 'float64', 'pressure mat server corner:off': 'float64', 'balcony door:off': 'float64', 'frame:off': 'float64', 'press bed right:off': 'float64', 'toaster:off': 'float64', 'microwave:off': 'float64', 'kitchen pir:off': 'float64'})
        ddf = dd.from_pandas(df_acts.copy(), npartitions=12)
        ddf = ddf.map_partitions(lambda dff: dff.apply(func, args=[raw], axis=1), meta=meta)
        ddf = ddf[:, ].drop(columns=[START_TIME, END_TIME])
        ddf = ddf.groupby(ACTIVITY).sum()
        ddf.visualize(filename='test.svg')
        df = ddf.compute()

    # Check the case when an activity does not match any device states
    return df.T



def contingency_table_states(df_devs, df_acts, other=False, n_jobs=1):
    """
    Compute the time a device is "on" or "off" respectively
    during the different activities.

    Parameters
    ----------
    df_devs : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    other : bool
        Determines whether gaps between activities should be assigned
        the activity *other* or be ignored.

    Examples
    --------
    >>> from pyadlml.stats import contingency_duration
    >>> contingency_duration(data.df_devs, data.df_activities)
    activity                     get drink ...             use toilet
    Hall-Bedroom door Off  0 days 00:01:54 ... 0 days 00:12:24.990000
    Hall-Bedroom door On   0 days 00:14:48 ... 0 days 03:02:49.984000
    ...                                ...
    Washingmachine On      0 days 00:00:00 ...        0 days 00:00:00
    [14 rows x 7 columns]

    Returns
    -------
    df : pd.DataFrame
    """
    TD = 'td'
    ON_praefix = 'on'
    OFF_praefix = 'off'
    SEP = ':'

    df_devs = df_devs.copy().reset_index(drop=True).sort_values(by=TIME)
    df_acts = df_acts.copy().reset_index(drop=True).sort_values(by=START_TIME)
    df_acts[TD] = df_acts[END_TIME] - df_acts[START_TIME]

    dtypes = infer_dtypes(df_devs)
    start_time = df_acts.iat[0, 0] - pd.Timedelta('1s')
    end_time = df_acts.iat[-1, 1] + pd.Timedelta('1s')
    from pyadlml.dataset._core.devices import device_events_to_states
    df_devs = device_events_to_states(df_devs, extrapolate_states=True,
                                      start_time=start_time, end_time=end_time)
    bool_mask_true = (df_devs[DEVICE].isin(dtypes[BOOL])) & (df_devs[VALUE] == True)
    bool_mask_false = (df_devs[DEVICE].isin(dtypes[BOOL])) & (df_devs[VALUE] == False)
    mask_cat = (df_devs[DEVICE].isin(dtypes[CAT]))

    df_devs.loc[bool_mask_true, DEVICE] = df_devs.loc[bool_mask_true, DEVICE] + SEP + ON_praefix
    df_devs.loc[bool_mask_false, DEVICE] = df_devs.loc[bool_mask_false, DEVICE] + SEP + OFF_praefix
    df_devs.loc[mask_cat, DEVICE] = df_devs.loc[mask_cat, DEVICE] + SEP + df_devs.loc[mask_cat, VALUE]
    df = df_devs.drop(columns=VALUE)

    for act in df_acts[ACTIVITY].unique():
        df[act] = pd.Timedelta('0ns')

    def func(row, df_acts):
        assert row.end_time - row.start_time > pd.Timedelta('0ns')
        mask_inside = (row.start_time <= df_acts[START_TIME]) & (df_acts[END_TIME] <= row.end_time)
        mask_right_ov = (df_acts[START_TIME] < row.end_time) & (row.end_time < df_acts[END_TIME])\
                      & (row.start_time < df_acts[START_TIME])
        mask_left_ov = (df_acts[START_TIME] < row.start_time) & (row.start_time < df_acts[END_TIME])\
                      & (df_acts[END_TIME] < row.end_time)
        mask_total_ov = (df_acts[START_TIME] <= row.start_time) & (row.end_time <= df_acts[END_TIME])

        overlap_ins = df_acts[mask_inside]
        overlap_left = df_acts[mask_left_ov]
        overlap_right = df_acts[mask_right_ov]
        overlap_total = df_acts[mask_total_ov]

        if not overlap_total.empty:
            assert overlap_ins.empty and overlap_left.empty and overlap_right.empty
            assert len(overlap_total) == 1, 'trouble double dup dup.'
            # One activity overlaps the whole state
            ol = overlap_total.iloc[0]
            row[ol[ACTIVITY]] = row.end_time - row.start_time
            return row

        if overlap_ins.empty and overlap_right.empty and overlap_left.empty:
            # No matching activity return row
            return row

        if not overlap_left.empty:
            assert len(overlap_left) == 1, 'trouble double dup dup.'
            ol = overlap_left.iloc[0]
            row[ol[ACTIVITY]] += ol[END_TIME] - row.start_time

        if not overlap_right.empty:
            assert len(overlap_right) == 1, 'trouble double dup dup.'
            ol = overlap_right.iloc[0]
            row[ol[ACTIVITY]] += row.end_time - ol[START_TIME]

        assert len(overlap_right) < 2 and len(overlap_left) < 2, 'Trouble trouble double dup.'

        if not overlap_ins.empty:
            ins_acts = overlap_ins[ACTIVITY].unique()
            state_agg = overlap_ins[[TD, ACTIVITY]].groupby(by=[ACTIVITY])[TD].sum()
            for act in ins_acts:
                row.at[act] += state_agg[act]

        return row

    if n_jobs > 1:
        import dask.dataframe as dd
        ddf = dd.from_pandas(df, npartitions=n_jobs)

        meta = {**{START_TIME: 'datetime64[ns]', END_TIME: 'timedelta64[ns]', DEVICE: 'object'}, 
                **{name: 'object' for name in df.columns[3:]}}

        ddf = ddf.map_partitions(lambda dff: dff.apply(func, args=[df_acts], axis=1), meta=meta)

        ddf = ddf.drop(columns=[START_TIME, END_TIME])
        ddf = ddf.groupby(DEVICE).sum()

        # Visualize as graph
        #ddf.visualize(filename='test.svg')
        df = ddf.compute()

    else:
        df = df.apply(func, args=[df_acts], axis=1)
        df = df.drop(columns=[START_TIME, END_TIME])
        df = df.groupby(DEVICE).sum()

    # Add row for categories for binary devices that are not present 
    devices_in_index = df.index.values
    zero_line = {col: pd.Timedelta('0ns') for col in df.columns}
    for dev in dtypes['boolean']:
        dev_on_name = dev + SEP + ON_praefix
        dev_off_name = dev + SEP + OFF_praefix
        if dev_on_name not in devices_in_index:
           df = pd.concat([df, pd.Series(name=dev_on_name, data=zero_line).to_frame().T], axis=0)
        if dev_off_name not in devices_in_index:
           df = pd.concat([df, pd.Series(name=dev_off_name, data=zero_line).to_frame().T], axis=0)
    
    return df


def mutual_info_states(df_devs, df_acts):
    raise NotImplementedError


def mutual_info_events(df_devs: pd.DataFrame, df_acts: pd.DataFrame, kind='pointwise', other=True) -> pd.DataFrame:
    """ Compute the mututal information between events and activities. How much does the 
        occurence, pattern of events from a certain device reduce the uncertainty in observing 
        some activity. This method provides

        - pointwise mutual information measures the co-occurence of events during activities 
          :math:`I(X,Y) = \log_2 \\frac{p(x,y)}{p(x)p(y)}` where :math:`p(x) = \\frac{o_x}{N}` 
          with :math:`o_x` being the event number of device :math:`x` and :math:`N` the total 
          number of events :math:`o_y` being the number of events labeled as an activity :math:`y` 
          and finally :math:`p(x,y) = \frac{o_{xy}}{N}`

          for further reading https://en.wikipedia.org/wiki/Pointwise_mutual_information.

        - event count mi 
        - latency of first event after activity onset
        - coarse resolution binary pattern represenation of device event activity

    Parameters
    ----------
    df_devs: pd.DataFrame
        TODO add desc
    df_acts: pd.DataFrame
        TODO add desc
    kind : str one of ['pointwise', 'timetofirstevent', '']

    Returns 
    -------

    """
    assert kind in ['pointwise', 'eventcount', 'timetofirstevent', 'binary_pattern']

    if kind == 'pointwise':
        c_xy_events = contingency_table_events(df_devs, df_acts)
        c_xy_events = c_xy_events.set_index('device')

        p_xy = c_xy_events.values
        N = p_xy.sum()
        p_y = np.tile(p_xy.sum(axis=0), (p_xy.shape[0], 1))/N
        p_x = np.tile(p_xy.sum(axis=1), (p_xy.shape[1], 1)).T/N
        p_xy = p_xy/N
        mi = np.log2(p_xy / np.multiply(p_x, p_y))

        return pd.DataFrame(data=mi, index=c_xy_events.index, columns=c_xy_events.columns)

    if kind == 'eventcount':
        """
        compute 
        """
        df = label_data(df_devs, df_acts, other=other)
        # Id each trial
        df['trial_id'] = (df[ACTIVITY].shift(1) != df[ACTIVITY]).astype(int)
        df_grp = df.groupby(by=[ACTIVITY, DEVICE, 'trial_id']).count()
        devs = df_devs[DEVICE].unique()
        acts = df_acts[ACTIVITY].unique()
        from itertools import product
        # Get the event count over all trials for each 
        event_count = pd.DataFrame(index=devs, columns=acts, dtype=object)
        for (dev, act) in product(dev, act):
            event_count.at[dev, act] = df_grp.loc[(act, dev)][VALUE].values

        print()
    if kind == 'chi2':
        c_xy_events = contingency_table_events(df_devs, df_acts)
        c_xy_events = c_xy_events.set_index('device')
        from sklearn.feature_selection import chi2
        c2 = chi2(c_xy_events.values) * (1/(2*np.log(2)))


    
    return mi



def time_dependent_firing_rate(df_devs, df_acts, device, activity, bin_width='10s'):
    """
    A trial is when an activity is performed.

    time t is measured w.r. to the start of the activity.
    The number of occurrences of events :math:`n_K(t;t+ \delta t)
    Same as the peri-stimulus time histogram.

    Parameters
    ----------

    bin_width : str one of ['xs', ]
        The bin width :math:`delta_t`


    """

    df_acts = df_acts[df_acts[ACTIVITY] == activity].sort_values(by=START_TIME)\
                                                    .reset_index(drop=True)
    df_devs = df_devs[df_devs[DEVICE] == device].sort_values(by=TIME)\
                                                .reset_index(drop=True)
    bin_width = pd.Timedelta(bin_width)
    
    n_bins = np.ceil(max((df_acts[END_TIME] - df_acts[START_TIME])/bin_width))
    max_dt = n_bins*bin_width

    # the number of occurrences of event n_K(t; t+\Delta t)
    counts = np.zeros(n_bins)

    # Number of trials
    K = df_acts.shape[0]

    for _, act_ser in df_acts.iterrows():

        dev_mask = (act_ser[START_TIME] < df_devs[TIME]) \
                    & (df_devs[TIME] < act_ser[END_TIME])
        devs = df_devs[dev_mask].copy() 

        if devs.empty:
            continue
        vals = (devs[TIME] - act_ser[START_TIME]).apply(_get_freq_func('minutes')).values

        add, bins = np.histogram(vals, bins=n_bins, range=(0,max_dt), density=False)
        counts += add

    # Compute event density as p(t) = 1/(\delta t) (n_k(t;t+\deltat))/K
    counts = (1/bin_width)*counts/K

    return counts, bins


import math
from collections import OrderedDict
import plotly.express as px
from pyadlml.dataset.stats.activities import _get_freq_func

from pyadlml.dataset.stats.devices import event_cross_correlogram3
from .util import dyn_event_marker_size, legend_current_items, _style_colorbar, dyn_y_label_size, remove_whitespace_around_fig
import plotly
import pandas as pd
import numpy as np

from plotly.subplots import make_subplots
import plotly.graph_objects as go
from plotly.express import IdentityMap
from plotly.express._core import apply_default_cascade, infer_config, \
    get_label, make_trace_kwargs, get_decorated_label, init_figure, make_trendline_spec, configure_axes, \
    configure_animation_controls, process_dataframe_timeline, one_group, \
    get_groups_and_orders

#from plotly.subplots import _subplot_type_for_trace_type, _set_trace_grid_reference
from plotly.validators.choropleth import ColorscaleValidator
from pyadlml.constants import DEVICE, TIME, ACTIVITY, START_TIME, END_TIME, VALUE, NUM, BOOL, \
    CAT, STRFTIME_DATE
import plotly.express as px

from pyadlml.dataset._core.devices import device_events_to_states
from pyadlml.dataset._core.activities import ActivityDict, create_empty_activity_df
from pyadlml.dataset.plot.plotly.util import format_device_labels
from pyadlml.dataset.plot.plotly.activities import _set_compact_title
from pyadlml.dataset.stats.acts_and_devs import contingency_table_states, contingency_table_events
from pyadlml.dataset.util import select_timespan, df_difference, activity_order_by, device_order_by, infer_dtypes


__all__ = ['activities_and_devices', 'contingency_states', 'contingency_events']

def _dynamic_and_height(nr_subj, nr_devs):
    # Additional rows for sel. marker
    m = (350 - 380)/(10 - 12)
    b = 0
    possible_rows = nr_subj + nr_devs + (nr_subj + 1)    
    height = int(m*possible_rows + b)
    print(f'height: {height} = {m}*{possible_rows} + {b}')
    return min(height, 1000)



def _plot_device_states_into_fig(fig: go.Figure, df_devs: pd.DataFrame,  df_devs_usel: pd.DataFrame,
                                 df_devs_outside: pd.DataFrame, dev_order: list, st=None, et=None,
                                 row=1, col=1) -> go.Figure:
    """
    Parameters
    ----------
    df_devs_outside : list of dicts
        Each dictionary contains the key 'df' and a 'color' and a 'opacity'


    """
    EVENT_COLOR = 'Black'
    COL_ON = 'teal'
    COL_OFF = 'lightgray'
    ON = 'on'
    OFF = 'off'
    #marker = dict(size=5, symbol=42, line=dict(color=EVENT_COLOR, width=1))

    df_devs = df_devs.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)
    devs = df_devs[DEVICE].unique()
    dtypes = infer_dtypes(df_devs)
    df_devs = device_events_to_states(df_devs, start_time=st, end_time=et,
                                      extrapolate_states=True).reset_index(drop=True)

    if df_devs_usel is not None:
        mark_selected = {}
        df_devs_usel = df_devs_usel.rename(columns={TIME: START_TIME})
        if END_TIME not in df_devs_usel.columns:
            tmp = df_devs.copy()
            comp_df = tmp[[START_TIME, DEVICE, VALUE]]\
                      .merge(df_devs_usel, indicator=True, how='left')
            mask = (comp_df['_merge'] == 'both')
            tmp2 = tmp[mask]
            df_devs_usel = tmp2
        df_devs_usel = _endtime_to_offset(df_devs_usel, replace=False)

        mask_bool = df_devs_usel[DEVICE].isin(dtypes[BOOL])
        df_devs_usel.loc[mask_bool, VALUE] = df_devs_usel.loc[mask_bool, VALUE].map({True: ON, False: OFF})

    # Create offset and color
    df_devs = _endtime_to_offset(df_devs, replace=False)
    mask_bool = df_devs[DEVICE].isin(dtypes[BOOL])
    df_devs.loc[mask_bool, VALUE] = df_devs.loc[mask_bool, VALUE].map({True: ON, False: OFF})

    # A mapping from device to data index
    #data_dict = {}
    j = 0
    #first_bool = True
    set_legendgroup_title = 'States'

    #def create_trace_boolean(df, state):
    #    nonlocal set_legendgroup_title

    #    df['diff'] = (df[END_TIME] - df[START_TIME]).astype(str)
    #    cd = df[['diff', END_TIME]].values
    #    # Add state information to custom_data
    #    vals = np.expand_dims(np.full(cd.shape[0], state), axis=1)
    #    cd = np.hstack([cd, vals])
    #    marker_color = COL_ON if state == ON else COL_OFF

    #    trace = go.Bar(name=state,
    #                    base=df[START_TIME],
    #                    meta=dev,
    #                    x=df['offset'],
    #                    y=df[DEVICE],
    #                    legendgrouptitle_text=set_legendgroup_title,
    #                    marker_color=marker_color,
    #                    customdata=cd,
    #                    legendgroup=state,
    #                    orientation='h',
    #                    width=0.3,
    #                    textposition='auto',
    #                    showlegend=first_bool,
    #                    hovertemplate=hover_template,
    #    )
    #    # The very first time create the states for the categorical legend group titles
    #    set_legendgroup_title = None 
    #    return trace

    def create_traces_categorical(df):
        """ df [TIME, DEVICE, VALUE] of only one device
        """
        nonlocal set_legendgroup_title
        nonlocal cat_idx
        nonlocal hover_template

        device_name = df[DEVICE].unique()[0]
        df = df.copy()\
               .sort_values(by=START_TIME)\
               .reset_index(drop=True)
        df[VALUE] = df[VALUE].astype(str)
        df['diff'] = (df[END_TIME] - df[START_TIME]).astype(str)



        categories = df[VALUE].unique()
        trace_lst = []
        for cat in categories:
            
            # TODO refactor, when categories of devices and activities overlap 
            # commented line below does not work
            #if cat not in legend_current_items(fig):
            if cat not in cat_col_map.keys():
                showlegend = True
                if cat not in cat_col_map.keys():
                    cat_col_map[cat] = cat_colors[cat_idx]
                    cat_idx +=1
            else:
                showlegend = False

            marker_color = cat_col_map[cat]
            mask_cat = (df[VALUE] == cat)

            cd = df.loc[mask_cat, [DEVICE, 'diff', END_TIME]].values
            vals = np.expand_dims(np.full(cd.shape[0], cat), axis=1)
            cd = np.hstack([cd, vals])

            trace = go.Bar(name=cat,
                           meta=dev,
                           base=df.loc[mask_cat, START_TIME],
                           x=df.loc[mask_cat, 'offset'],
                           y=df.loc[mask_cat, DEVICE],
                           legendgrouptitle_text=set_legendgroup_title,
                           marker_color=marker_color,
                           legendgroup=cat,
                           customdata=cd,
                           orientation='h',
                           width=0.3,
                           textposition='auto',
                           showlegend=showlegend,
                           hovertemplate=hover_template,
            )
            trace_lst.append(trace)
            set_legendgroup_title = None

        return trace_lst


    cat_colors = px.colors.qualitative.Pastel1 \
               + px.colors.qualitative.Pastel2
    cat_col_map = {ON:COL_ON, OFF:COL_OFF}
    cat_idx = 0
    hover_template = '%{customdata[0]}<br>' + \
                        'Start_time: %{base|' + STRFTIME_DATE + '}<br>' + \
                        'End_time: %{customdata[2]|' + STRFTIME_DATE + '}<br>' + \
                        'Duration: %{customdata[1]}<br>' + \
                        'State: %{customdata[3]}<extra></extra>'

    for i, dev in enumerate(devs):
        df = df_devs[df_devs[DEVICE] == dev].copy()

        if dev in dtypes[BOOL]:
            df_on = df[df[VALUE] == ON]
            df_off = df[df[VALUE] == OFF]

            if df_devs_usel is not None:
                mark_selected[dev] = {}
                comp_df = df_on.copy().merge(df_devs_usel, indicator=True, how='left')
                tmp = np.where((comp_df['_merge'] == 'both').values)[0]
                comp_df2 = df_off.copy().merge(df_devs_usel, indicator=True, how='left')
                tmp2 = np.where((comp_df2['_merge'] == 'both').values)[0]
                mark_selected[dev][ON] = tmp
                mark_selected[dev][OFF] = tmp2

            traces = create_traces_categorical(df)
            fig.add_traces(traces)
            #trace_on = create_trace_boolean(df_on, ON)
            #trace_off = create_trace_boolean(df_off, OFF)
            #data_dict[dev] = [len(fig.data), len(fig.data)+1]
            #fig.add_traces([trace_on, trace_off])

            #first_bool = False
        elif dev in dtypes[CAT]:
            traces = create_traces_categorical(df)
            fig.add_traces(traces)
            j += 1

        #elif dev in dtypes[NUM]:
        elif False:
            values = pd.to_numeric(df[VALUE])
            # [min, max] scaling -> [0, 1]
            values_norm = (values-values.min())/(values.max()-values.min())
            #values_norm = values_norm + i - 0.25
            print(values_norm)
            trace = go.Scatter(
                name=dev,
                meta=dev,
                mode='lines',
                x=df[START_TIME],
                y=values_norm,
            )
            trace._subplot_row = 1
            trace._subplot_col = 1
            

            fig = make_subplots(
                #rows=1,
                #cols=1,
                row_heights=[1.0],
                horizontal_spacing=0.02,
                vertical_spacing=0.03,
                shared_xaxes='all',
                shared_yaxes='all',
                column_widths=[1.0],
                specs=[[{"type":"xy", "secondary_y": True}]], 
                figure=fig,
            )
            #fig = make_subplots(

            #    shared_xaxes='all',
            #    shared_yaxes='all',
            #    start_cell="bottom-left",
            #    horizontal_spacing=0.02,
            #    vertical_spacing=0.03,
            #    subplot_titles=[],
            #    column_widths=[1.0],
            #    row_heights=[1.0],
            #    specs=[[{'type': 'xy'}]],
            #    figure=fig)
            #trace.update({'xaxis': 'x', 'yaxis': 'y'})
            fig.add_trace(go.Bar(
                            name=dev,
                            base=df[START_TIME],
                            x=df['offset'],
                            y=df[DEVICE],
                            orientation='h'
            ))
            fig.add_trace(trace, secondary_y=True)
            fig.update_yaxes(range=[0, len(devs)], secondary_y=True) 


        # Create user_selection for each trace
        if df_devs_usel is not None:
            if dev in (dtypes[BOOL] + dtypes[CAT]): 
                for cat in df[VALUE].unique():
                    fig.update_traces(selector=dict(meta=dev, name=cat),
                            selectedpoints=mark_selected[dev][cat],
                            selected={'marker': {'opacity': 1.0, 'color': cat_col_map[cat]}},
                            unselected={'marker': {'opacity': 0.3, 'color': cat_col_map[cat]}})
                # TODO not for, flag for deletion 
                #fig.update_traces(selector=dict(meta=dev, name=OFF),
                #           selectedpoints=mark_selected[dev][OFF],
                #           selected={'marker': {'opacity': 1.0, 'color': COL_OFF}},
                #           unselected={'marker': {'opacity': 0.3, 'color': COL_OFF}})
                #fig.update_traces(selector=dict(meta=dev, name=ON),
                #           selectedpoints=mark_selected[dev][ON],
                #           selected={'marker': {'opacity': 1.0, 'color': COL_ON}},
                #           unselected={'marker': {'opacity': 0.3, 'color': COL_ON}})
            else:
                raise NotImplementedError


    fig.update_layout(yaxis_type='category')
    fig.update_yaxes(categoryorder='array', categoryarray=np.flip(dev_order))
    return fig

def _abbreviate_long_names(arr: np.ndarray) -> np.ndarray:
    THRESHOLD = 10
    def func(x):
        if len(x) > THRESHOLD:
            # REmove preafix i.e. binary_sensor.blablbla
            x = "".join([s for s in x.split('.')[0]])
        if len(x) > THRESHOLD:
            x = "".join([s[:3] for s in x.split('_')])
        if len(x) > THRESHOLD:
            x = x[:THRESHOLD]
        else:
            return x
    vf = np.vectorize(func)
    arr = vf(arr) 
    return arr


def _plot_device_events_into_fig(fig: go.Figure, df_devs: pd.DataFrame,  df_devs_usel: pd.DataFrame,
                                 df_devs_outside: pd.DataFrame, dev_order: list,
                                 marker_height=5,
                                 ) -> go.Figure:
    """
    Parameters
    ----------
    df_devs_outside : list of dicts
        Each dictionary contains the key 'df' and a 'color' and a 'opacity'


    """
    # Enable webgl rendering
    scatter = go.Scattergl if len(df_devs) > 15000 else go.Scatter
    EVENT_COLOR = 'Black'
    hover_template = '%{customdata[1]}<br>%{x|' + STRFTIME_DATE + '}<br>Event: %{customdata[0]} <extra></extra>'


    marker = dict(size=marker_height, symbol=42, line=dict(color=EVENT_COLOR, width=1))
    fig.update_layout(yaxis_type='category')
    fig.update_yaxes(categoryorder='array', categoryarray=np.flip(dev_order))
    fig.add_trace(scatter(
        mode='markers', y=df_devs[DEVICE], x=df_devs[TIME],
        customdata=df_devs[[VALUE, DEVICE]].values,
        hovertemplate=hover_template,
        showlegend=False, marker=marker))

    if not df_devs_outside.empty:
        marker['opacity'] = 0.1
        marker['line']['color'] = 'Grey'
        fig.add_trace(scatter(mode='markers', y=df_devs_outside[DEVICE],
                                 x=df_devs_outside[TIME], showlegend=False,
                                 marker=marker, hoverinfo='skip')
                      )

    # Create user_selection for each trace
    if df_devs_usel is not None:
        df_devs_usel = df_devs_usel[[TIME, DEVICE, VALUE]]
        # Get the indices in the trace where
        comp_df = df_devs.copy().merge(df_devs_usel, indicator=True, how='left')
        mark_selected = np.where((comp_df['_merge'] == 'both').values)[0]
        unselected = {'marker': {'opacity': 0.2, 'color': EVENT_COLOR}}
        selected = {'marker': {'opacity': 1.0, 'color': 'Red'}}
        fig.update_traces(selector=dict(type="scatter"),
                          selectedpoints=mark_selected,
                          selected=selected, unselected=unselected)

    return fig


def _determine_start_and_end(df_acts: dict, df_devs: pd.DataFrame, st: pd.Timestamp, et: pd.Timestamp):
    """ Determine the start and endpoint with regard to optional given parameters
    """
    if isinstance(df_acts, pd.DataFrame):
        act_st = df_acts[START_TIME].iloc[0]
        act_et = df_acts[END_TIME].iloc[-1]
    elif isinstance(df_acts, ActivityDict):
        act_st = df_acts.min_starttime()
        act_et = df_acts.max_endtime()

    data_st = min(df_devs[TIME].iloc[0], act_st) - pd.Timedelta('1ms')
    data_et = max(df_devs[TIME].iloc[-1], act_et) + pd.Timedelta('1ms')
    if st is None:
        draw_start_line = False
        st = data_st
    else:
        # Draw the line only if the given start is the right-most
        draw_start_line = (max(st, data_st) == st)
        st = max(st, data_st)
    if et is None:
        draw_end_line = False
        et = data_et
    else:
        # Draw the line only if the given end is the left-most
        draw_end_line = (min(et, data_et) == et)
        et = min(et, data_et)
    return st, et, draw_start_line, draw_end_line


def act_difference(df_acts_sel, df_acts, st, et):
    """ Get outside activities with enveloping activities correctly clipped

    Parameters
    ----------
    df_acts_sel : pd.DataFrame or ActivityDict or List
        asdf
    df_acts : pd.DataFrame or ActivityDict or List
    st : str
    et : str

    Returns
    -------

    """
    acts_inst = type(df_acts)
    dct_acts = ActivityDict.wrap(df_acts)
    dct_acts_sel = ActivityDict.wrap(df_acts_sel)
    dct_acts_outside = ActivityDict()

    # Check if all keys are equal
    assert set(dct_acts_sel.keys()) == set(dct_acts.keys())

    for key in dct_acts.keys():
        df_acts = dct_acts[key]
        df_acts_sel = dct_acts_sel[key]

        df_acts_outside = df_difference(df_acts_sel, df_acts)\
                        .sort_values(by=START_TIME)\
                        .reset_index(drop=True)

        # ao contains both the cut up and the non-cut activities, adjust the start activity
        if not df_acts_outside.empty:
            df_acts_outside[ACTIVITY] = 'not selected'

            # Both the old time and the new start_time are in df_acts_outside
            # Get entry of outside corresponding to start_time
            act_st_split_et = df_acts_outside[df_acts_outside[START_TIME] == st]
            if not act_st_split_et.empty:
                idxs_acts_st_split = df_acts_outside[(df_acts_outside[END_TIME] == act_st_split_et.iat[0, 1])].index
                df_acts_outside.iat[idxs_acts_st_split[0], 1] = st - pd.Timedelta('1ms')
                df_acts_outside = df_acts_outside.drop(index=idxs_acts_st_split[1])

                df_acts_outside = df_acts_outside.reset_index(drop=True)


            act_st_split_st = df_acts_outside[df_acts_outside[END_TIME] == et]
            if not act_st_split_st.empty:
                idxs_acts_et_split = df_acts_outside[(df_acts_outside[START_TIME] == act_st_split_st.iat[0, 0])].index
                df_acts_outside.iat[idxs_acts_et_split[1], 0] = et - pd.Timedelta('1ms')
                df_acts_outside = df_acts_outside.drop(index=idxs_acts_et_split[0])

                df_acts_outside = df_acts_outside.reset_index(drop=True)

        dct_acts_outside[key] = df_acts_outside

    return dct_acts_outside.unwrap(inst_type=acts_inst)


def _plot_selected_activity_marker(fig, df, label='Selected Activity mark'):
    """ Plot markers at the midpoint of selected activities
    """
    df = df[[START_TIME, END_TIME, ACTIVITY]].copy()
    cd = df.values
    diff = df[END_TIME] - df[START_TIME]
    df['mid_point'] = df[START_TIME] + diff/2
    y_label = label
    df['y'] = y_label

    marker = dict(size=5, symbol=5, line=dict(color='Red', width=1))
    hover_template = 'Activity: %{customdata[2]}<br>' + \
                     'Start_time: %{customdata[0]|' + STRFTIME_DATE + '}<br>' + \
                     'End_time: %{customdata[1]|' + STRFTIME_DATE + '}<br>' + \
                     '<extra></extra>'

    fig.add_trace(go.Scatter(
        mode='markers', y=df['y'], x=df['mid_point'],
        customdata=cd,
        marker=marker,
        hovertemplate=hover_template,
        showlegend=False))
    y_axis_order = [y_label] + fig.layout.yaxis.categoryarray.tolist()
    fig.update_yaxes(categoryarray=y_axis_order)

    return fig


def _plot_selected_device_marker(fig, df, df_devs, states=False, label='Sel devices'):
    df = df[[TIME, DEVICE, VALUE]].copy()
    df_devs = df_devs[[TIME, DEVICE, VALUE]].copy()\
        .sort_values(by=TIME)\
        .reset_index(drop=True)

    y_label = label
    df['y'] = y_label

    if states:
        # Set the categorical and boolean markers to midpoint of state
        dtypes = infer_dtypes(df)
        df[END_TIME] = df[TIME].copy()
        for dev in dtypes[BOOL] + dtypes[CAT]:
            df_tmp = df_devs[df_devs[DEVICE] == dev].copy()
            df_tmp[END_TIME] = df_tmp[TIME].shift(-1)
            mask = (dev == df[DEVICE])
            times = df.loc[mask, TIME]
            vals = df_tmp[df_tmp[TIME].isin(times)]
            df.loc[mask, END_TIME] = vals[END_TIME]

        df[START_TIME] = df[TIME]
        df[TIME] = df[TIME] + (df[END_TIME] - df[TIME])/2
        custom_data = df[[DEVICE, START_TIME, END_TIME]].values
        hover_template = 'Device: %{customdata[0]}<br>' + \
                         'Start_time: %{customdata[1]|' + STRFTIME_DATE + '}<br>' + \
                         'End_time: %{customdata[2]|' + STRFTIME_DATE + '}<br>' + \
                         '<extra></extra>'
    else:
        custom_data = df[DEVICE]
        hover_template = 'Device: %{customdata}<br>Time: %{x}<extra></extra>'
    marker = dict(size=5, symbol=5, line=dict(color='Red', width=1))
    fig.add_trace(go.Scatter(
        mode='markers', y=df['y'], x=df[TIME],
        customdata=custom_data,
        marker=marker,
        hovertemplate=hover_template,
        showlegend=False))
    y_axis_order = [y_label, *list(fig.layout.yaxis.categoryarray)]
    fig.update_yaxes(categoryarray=y_axis_order)
    return fig



def activities_and_devices(df_devs, df_acts, states=False, st=None, et=None,
                           act_order='alphabetical', dev_order='alphabetical', dev2area=None,
                           df_acts_usel=None, df_devs_usel=None, devs_usel_state=None,
                           height=None, zoomed_range=[], fig=None, row=None, col=None
                           ):
    """

    Parameters
    ----------
    df_acts : pd.Dataframe or List of dataframes or dict of  
    """

    st = pd.Timestamp(st) if st is not None else st
    et = pd.Timestamp(et) if et is not None else et

    # Determines if device markers indicate the event or a states midpoint
    if states:
        devs_usel_state = states if devs_usel_state is None else devs_usel_state
    else:
        devs_usel_state = False
    assert isinstance(devs_usel_state, bool), 'devs_usel_state has to be set with either a boolean or None'

    dct_acts = ActivityDict.wrap(df_acts)
    if df_acts_usel is not None:
        dct_acts_usel = ActivityDict.wrap(df_acts_usel)
    else:
        dct_acts_usel = None


    dct_acts = dct_acts.copy()
    df_devs = df_devs.copy()

 

    # Determine the start and endpoint with regard to optional given parameters
    st, et, draw_start_line, draw_end_line = _determine_start_and_end(dct_acts, df_devs, st, et)

    # Select the active displayed parts
    df_devs_sel, dct_acts_sel = select_timespan(df_devs, dct_acts, st, et, clip_activities=True)

    # Get the y-axis and label order
    act_order = activity_order_by(dct_acts_sel.concat(), rule=act_order)
    dev_order = device_order_by(df_devs_sel, rule=dev_order, dev2area=dev2area)


    # determinte visual properties
    nr_devs = len(dev_order)
    nr_subjs = len(dct_acts)
    height = _dynamic_and_height(nr_devs, nr_subjs) if height is None else height
    marker_height = dyn_event_marker_size(height, nr_devs)
    y_label_size = dyn_y_label_size(height, nr_devs)

    # Reconstruct outside parts
    df_devs_outside = df_difference(df_devs_sel, df_devs)
    dct_acts_outside = act_difference(dct_acts_sel, dct_acts, st, et)

    title = 'Activities and device events'
    if fig is None:
        col, row = 1,1
        fig = make_subplots(rows=row, cols=col)
    else:
        row = 1 if row is None else row
        col = 1 if col is None else col


    for key in reversed(list(dct_acts.keys())):
        if df_acts_usel is not None:
            try:
                usel = dct_acts_usel[key]
            except: 
                usel = None
        else:
            usel = None
        fig = _plot_activities_into_fig(fig, dct_acts_sel[key], usel, 
                                        dct_acts_outside[key], act_order,
                                        y_label=f'Acts: {key}'
                                        )


    if states:
        fig = _plot_device_states_into_fig(fig, df_devs_sel, df_devs_usel,
                                           df_devs_outside, dev_order, st, et, row, col)
    else:
        fig = _plot_device_events_into_fig(fig, df_devs_sel, df_devs_usel, df_devs_outside, 
                                           dev_order, marker_height)



    # Plot markers into trace
    if dct_acts_usel is not None:
        for key, usel in dct_acts_usel.items():
            fig = _plot_selected_activity_marker(fig, usel, label=f'Sel Act: {key}')

    if df_devs_usel is not None:
        fig = _plot_selected_device_marker(fig, df_devs_usel, df_devs, devs_usel_state)



    if draw_start_line:
        fig.add_vline(x=st, line_width=1, line_color="Grey")
    if draw_end_line:
        fig.add_vline(x=et, line_width=1, line_color="Grey")

    _set_compact_title(fig, title)
    fig.update_yaxes(title=None, fixedrange=True, tickfont=dict(size=y_label_size),
                     ticklabeloverflow='allow'
                     )
    if zoomed_range:
        zr_st = pd.Timestamp(zoomed_range[0])
        zr_et = pd.Timestamp(zoomed_range[1])
        fig.update_xaxes(type='date', range=[zr_st, zr_et])


    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    return fig


def _endtime_to_offset(df_act, replace=True):
    """Compute the end_time as numerical offset in ms"""
    x_start = pd.to_datetime(df_act[START_TIME])
    x_end = pd.to_datetime(df_act[END_TIME])
    col_label = END_TIME if replace else 'offset'
    df_act[col_label] = (x_end - x_start).astype("timedelta64[ms]")
    return df_act


def _plot_activities_into_fig(fig, df_acts: pd.DataFrame, df_acts_usel: pd.DataFrame,
                              df_acts_outside: pd.DataFrame, act_order: list, y_label='Activity',
                              color_discrete_map=None) -> go.Figure:
    """ Manually plot timeline from plotly.express.make_figure
        fig = px.timeline(df_acts, x_start=START_TIME, x_end=END_TIME, y='y_label',
                              color=ACTIVITY)

    Parameters
    ----------
    fig : obj.Figure

    df_acts : pd.DataFrame

    mask_unselected : pd.Series
        A mask that indicates the datapoints whichs opacity should be lowered.
    """
    def _helper_1(df):
        df = df.sort_values(by=START_TIME).reset_index(drop=True)
        df['y_label'] = y_label
        df = df[['y_label', START_TIME, END_TIME, ACTIVITY]]
        df['dur'] = (df[END_TIME] - df[START_TIME]).astype(str)
        return df


    df_acts = _helper_1(df_acts)
    if df_acts_outside is None:
        df_acts_outside = create_empty_activity_df()
    else:
        df_acts_outside = _helper_1(df_acts_outside)

    if df_acts_usel is not None:
        df_acts_usel = _helper_1(df_acts_usel)
        df_acts_usel = _endtime_to_offset(df_acts_usel)

    args = dict(data_frame=df_acts,
                x_start=START_TIME,
                x_end=END_TIME,
                y='y_label',
                color=ACTIVITY,
                color_discrete_map=color_discrete_map,
                category_orders={'activity': []},
                #categoryarray=category_orders,

                # Necessary to work
                template=None,
                color_discrete_sequence=None,
                hover_data=None,
    )

    trace_patch = dict(textposition="auto", orientation="h")
    layout_patch = dict(barmode="overlay")

    # Process args
    apply_default_cascade(args)
    args = process_dataframe_timeline(args)

    trace_specs, grouped_mappings, sizeref, _ = infer_config(
        args, go.Bar, trace_patch, layout_patch
    )

    grouper = [x.grouper or one_group for x in grouped_mappings] or [one_group]
    grouped = args["data_frame"].groupby(grouper, sort=False)

    # Grouped is already ordered TODO debug
    _, orders = get_groups_and_orders(args, grouper)
    #orders, sorted_group_names = get_orderings(args, grouper, grouped)

    # Assign colors to the groups
    for val in act_order:
        m = grouped_mappings[0]
        if val not in m.val_map:
            m.val_map[val] = m.sequence[len(m.val_map) % len(m.sequence)]


    set_legendgrouptitle = 'Activities'
    trace_lst = []
    for act_name in act_order:
        if not (df_acts[ACTIVITY] == act_name).any():
            continue

        # Get dataframe with of only activity
        df_sel_act = grouped.get_group(act_name)

        # Create the trace
        # TODO showlegend
        showlegend = act_name not in legend_current_items(fig)
        m = grouped_mappings[0]
        trace_color = m.val_map[act_name]

        hover_template = '<b>' + act_name + '</b><br>'\
                         + 'Start_time: %{base|' + STRFTIME_DATE + '}<br>' \
                         + 'End_time: %{x| ' + STRFTIME_DATE + '}' \
                         + '<br>Duration: %{customdata}<extra></extra>'

        trace = go.Bar(name=act_name,
                       legendgroup=act_name, 
                       showlegend=showlegend,
                       legendgrouptitle_text=set_legendgrouptitle,
                       marker_color=trace_color,
                       alignmentgroup=True, 
                       offsetgroup=act_name,
                       textposition='auto', 
                       orientation='h',
                       base=df_sel_act[START_TIME],
                       x=df_sel_act[END_TIME],
                       y=df_sel_act['y_label'],
                       customdata=df_sel_act['dur'],
                       hovertemplate=hover_template
        )
        set_legendgrouptitle = None

        trace_lst.append(trace)

        # Create user_selection for each trace
        if df_acts_usel is not None:
            # Get the indices in the trace where
            comp_df = df_sel_act.copy().merge(df_acts_usel, indicator=True, how='left')
            mark_selected = np.where((comp_df['_merge'] == 'both').values)[0]
            unselected = {'marker': {'opacity': 0.3, 'color': trace_color}}
            selected = {'marker': {'opacity': 1.0, 'color': trace_color}}
            trace.update(selectedpoints=mark_selected, selected=selected, unselected=unselected)

    if not df_acts_outside.empty:
        # Convert to bar
        act_name = 'not selected'
        df_sel_act = _endtime_to_offset(df_acts_outside)

        # Create the trace
        trace = go.Bar(name=act_name,
                       base=df_sel_act[START_TIME],
                       hoverinfo='skip',
                       marker=dict(color='rgba(58, 71, 80, 0.6)', opacity=0.2),
                       x=df_sel_act[END_TIME],
                       y=df_sel_act['y_label'],
                       customdata=df_sel_act['dur'],
                       textposition='auto',
                       orientation='h',
                       legendgroup=act_name,
                       showlegend=False,
                       alignmentgroup=True,
                       offsetgroup=act_name,
        )
        trace._subplot_row = 1
        trace._subplot_col = 1
        trace_lst.append(trace)

    fig = make_subplots(
        rows=1,
        cols=1,
        shared_xaxes='all',
        shared_yaxes='all',
        start_cell="bottom-left",
        horizontal_spacing=0.02,
        vertical_spacing=0.03,
        subplot_titles=[],
        column_widths=[1.0],
        row_heights=[1.0],
        specs=[[{'type': 'xy'}]],
        figure=fig)


    # Add traces, layout and frames to figure
    fig.add_traces(trace_lst)

    # Resizes bars to full size
    fig.update_layout({'barmode': 'overlay',
                       'legend': {'tracegroupgap': 0}
                       })
    fig.update_xaxes(type="date")

    return fig




def contingency_events(df_devs=None, df_acts=None, con_tab=None, scale='linear', height=350,
                       act_order='alphabetical', dev_order='alphabetical', n_jobs=1,
                       ) -> plotly.graph_objects.Figure:
    """
    """
    title = 'Activities vs device events'
    cbarlabel = 'count' if scale == 'linear' else 'log count'

    # get the list of cross tabulations per t_window
    if con_tab is not None:
        df_con = con_tab.copy()    # :pd.DataFrame
    else:
        df_con = contingency_table_events(df_devs, df_acts, n_jobs=n_jobs)

    df_con = df_con.set_index('device')

    act_order = activity_order_by(df_acts, rule=act_order)
    dev_order = device_order_by(df_devs, rule=dev_order)
    df_con = df_con[np.flip(act_order)]     # change order of rows
    df_con = df_con.reindex(dev_order)      # change order of columns

    # Extract values
    devs = df_con.index.values
    acts = df_con.columns.values
    vals = df_con.values
    z = vals if scale == 'linear' else np.log(vals)

    fig = go.Figure(data=go.Heatmap(
        z=z.T,
        x=devs,
        y=acts,
        customdata=vals.T,
        hovertemplate='"%{x}" during "%{y}" %{customdata} times<extra></extra>',
        colorscale='Viridis',
        hoverongaps=False))

    _set_compact_title(fig, title)

    fig.update_xaxes(tickangle=45)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _style_colorbar(fig, cbarlabel)

    return fig


def contingency_states(df_devs=None, df_acts=None, con_tab=None, scale='linear', height=350,
                       act_order='alphabetical', dev_order='alphabetical', n_jobs=1
                       ) -> plotly.graph_objects.Figure:
    """
    """
    title = 'Activities vs device states'
    cbarlabel = 'seconds' if scale == 'linear' else 'log seconds'

    # get the list of cross tabulations per t_window
    if con_tab is not None:
        df_con = con_tab.copy()    # :pd.DataFrame
        df_con.index.name = DEVICE
    else:
        df_con = contingency_table_states(df_devs, df_acts, n_jobs=n_jobs)

    # Determine device and activity labels order
    act_order = activity_order_by(df_acts, rule=act_order)
    df_con = df_con[np.flip(act_order)]     # change order of rows
    devs = df_con.index.values

    # Reorder and format devices
    dtypes = infer_dtypes(df_devs)
    dtypes.pop(NUM, None)
    new_devs, new_order = format_device_labels(devs, dtypes, boolean_state=True, categorical_state=True)
    devs = list(new_devs)
    vals = df_con.copy()\
        .reset_index()\
        .reindex(new_order)\
        .set_index(DEVICE)

    # Extract values
    acts = df_con.columns.values


    # Convert timedelta to nanoseconds
    z = vals.astype('timedelta64[ns]')/np.timedelta64(1, 'ns')
    z = z if scale == 'linear' else np.log(z)
    z = z.T

    # Create hoverdata, the duration as strings in (D,A,0) and full X names in (D,A,1)
    vals = vals.astype(str)
    tmp = np.tile(vals.index, (len(vals.columns), 1))
    cd = np.array([tmp, vals.values.T])
    cd = np.moveaxis(cd, 0, -1)


    # TODO hack, add whitespaces to similar categories to make them unique
    #      strings since otherwise plotly merges them into one category
    j = 0
    for i in range(len(devs)):
        if devs[i] == 'on':
            devs[i] = devs[i] + ' '*j
            j += 1

    fig = go.Figure(data=go.Heatmap(
        z=z,
        x=devs,
        y=acts,
        customdata=cd,
        hovertemplate='"%{customdata[0]}" during "%{y}" for %{customdata[1]}<extra></extra>',
        colorscale='Viridis',
        hoverongaps=False))

    _set_compact_title(fig, title)

    fig.update_xaxes(tickangle=45)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _style_colorbar(fig, cbarlabel)

    return fig


def event_correlogram(df_devs: pd.DataFrame, df_acts: pd.DataFrame, fix=[], to=[], maxlag='2min', binsize='2s', use_dask=False):

    from pyadlml.dataset.plot.plotly.devices import plotly_event_correlogram

    # Wrap if not list
    fix = [fix] if isinstance(fix, str) else fix
    to = [to] if isinstance(to, str) else to

    activities = df_acts[ACTIVITY].unique().tolist()
    if fix:
        assert np.array([e in activities for e in fix]).all()
        activities = fix
    
    devices = df_devs[DEVICE].unique().tolist()
    if to:
        assert np.array([e in devices for e in to]).all()
        devices = to

    df_acts[VALUE] = 1
    df_acts = df_acts.rename(columns={ACTIVITY: DEVICE})

    df_act_st = df_acts[[START_TIME, DEVICE, VALUE]].copy()
    df_act_st = df_act_st.rename(columns={START_TIME:TIME})
    df_act_et = df_acts[[END_TIME, DEVICE, VALUE]].copy()
    df_act_et = df_act_et.rename(columns={END_TIME:TIME})

    df_comb_st = pd.concat([df_devs, df_act_st], axis=0)\
                    .sort_values(by=TIME)\
                    .reset_index(drop=True)
    df_comb_et = pd.concat([df_devs, df_act_et], axis=0)\
                    .sort_values(by=TIME)\
                    .reset_index(drop=True)

    cc_st, bins, rows, cols = event_cross_correlogram3(
        df_comb_st, fix=activities, to=devices, maxlag=maxlag, binsize=binsize,
        use_dask=use_dask
    )
    cc_et, bins, rows, cols = event_cross_correlogram3(
        df_comb_et, fix=activities, to=devices, maxlag=maxlag, binsize=binsize,
        use_dask=use_dask
    )
    mid_point = int(np.floor(len(bins)/2))
    for i in range(len(devices)):
        for j in range(len(activities)):
            # Overwrite right half of st not including the odd zero bin
            cc_st[i,j, mid_point+1:] = cc_et[i,j,mid_point+1:]
            cc_st[i,j, mid_point] += cc_et[i,j, mid_point]

    return plotly_event_correlogram(
        cc_data=[cc_st, bins, rows, cols]
    )



@remove_whitespace_around_fig
def activity_vs_device_events_hist(df_devs, df_acts, device, activity, n_bins=20, normalize=False, height=600):
    fig = make_subplots(rows=2, cols=1, shared_xaxes=True, row_heights=[0.9, 0.1],
                        vertical_spacing=0.01, 
    
    )

    df_acts = df_acts[df_acts[ACTIVITY] == activity].sort_values(by=START_TIME)\
                                                    .reset_index(drop=True)
    df_devs = df_devs[df_devs[DEVICE] == device].sort_values(by=TIME)\
                                                .reset_index(drop=True)

    counts = np.zeros((n_bins,))

    if not normalize:
        df_acts['dt'] = df_acts[END_TIME] - df_acts[START_TIME]
        df_acts['minutes'] = df_acts['dt'].apply(_get_freq_func('minutes'))
        total_dt = max(df_acts['minutes'])
    else:
        total_dt = 1


    for _, act_ser in df_acts.iterrows():
        if normalize:
            dt2zeroone = lambda x : (x - act_ser[START_TIME])/(act_ser[END_TIME] - act_ser[START_TIME])

        dev_mask = (act_ser[START_TIME] < df_devs[TIME]) \
                    & (df_devs[TIME] < act_ser[END_TIME])
        devs = df_devs[dev_mask].copy() 

        if devs.empty:
            continue

        if normalize:
            vals = devs[TIME].apply(dt2zeroone).values
        else:
            vals = (devs[TIME] - act_ser[START_TIME]).apply(_get_freq_func('minutes')).values

        add, bins = np.histogram(vals, bins=n_bins, range=(0,total_dt), density=False)
        counts += add


    if not 'bins' in locals():
        # Case when no device event happened during given activities
        bins = np.linspace(0, total_dt, n_bins)
        print('Warning! No device event happend during activity')


    # Plot histogramm 
    # Move bars into intervals
    bar_width = (bins[-1] - bins[-2])*.8
    midpoints = bins[:-1] + (bins[1:] - bins[:-1])/2
    customdata = np.array([f'({b[0]:.3f}, {b[1]:.3f}]' for b in zip(bins[:-1], bins[1:])])


    fig.add_trace(
        go.Bar(
            x=midpoints,
            y=counts,
            name=device,
            width=bar_width,
            customdata=customdata,
            hovertemplate='%{customdata} %{y} times <extra></extra>',
            opacity=1,
            showlegend=True,
            marker=dict(line_width=0)
        ),row=1, col=1
    )
    fig.update_yaxes(title_text='counts', row=1, col=1)

    if normalize:
        fig.add_trace(
            go.Bar(
                x=[0,1],
                y=[activity]*2,
                orientation='h',
                opacity=1,
                showlegend=False,
                marker=dict(line_width=0)
            ),row=2, col=1
        )
    else:
        fig.add_trace(
            go.Box(
                name=activity,
                x=df_acts['minutes'],
            ),row=2, col=1
        )
    if normalize:
        fig.update_xaxes(range=[0,1], title_text='normalized range', row=2, col=1)
    else:
        fig.update_xaxes(title_text='minutes', row=2, col=1)

    fig.update_layout(title_text="Device events over one activity", height=height)

    return fig


import sys
import pathlib

from testing.test_preprocessing import TestPreprocessingBase

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from test_dataset_base import TestDatasetBase


from pyadlml.dataset import set_data_home
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, TIME,\
    set_data_home, fetch_mitlab, clear_data_home
TEST_DATA_HOME = '/tmp/pyadlml_testing'
MITLAB_DS_1 = 'subject1'
MITLAB_DS_2 = 'subject2'

class TestDatasetMitlabSubject1(TestDatasetBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_mitlab
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.num_activities = [22]
        self.len_activities = [296]
        self.num_rec_acts = [22]

        self.len_devices = 5196
        self.num_devices = 72
        self.num_rec_devs = 72
        self.num_rec_binary_devs = 72

    def _setUp(self):
        self.data = fetch_mitlab(keep_original=True, cache=False, subject='subject1')


class TestDatsetMitlabSubject2(TestDatasetBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_mitlab
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']
        self.num_activities = [24]
        self.len_activities = [219]
        self.num_rec_acts = [24]

        self.len_devices = 3198
        self.num_devices = 68
        self.num_rec_devs = 68
        self.num_rec_binary_devs = 68

    def _setUp(self):
        self.data = fetch_mitlab(keep_original=True, cache=False, subject='subject2')


class TestDatasetMitlabSubject1Preprocessing(TestPreprocessingBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_mitlab
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']

    def _setUp(self):
        self.data = fetch_mitlab(cache=True, subject='subject1')


class TestDatasetMitlabSubject2Preprocessing(TestPreprocessingBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_mitlab
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']

    def _setUp(self):
        self.data = fetch_mitlab(cache=True, subject='subject2')

if __name__ == '__main__':
    unittest.main()


from pathlib import Path

from sklearn.tree import export_graphviz
from pyadlml.dataset import *
from pyadlml.preprocessing import StateVectorEncoder, LabelMatcher, DropTimeIndex, DropDuplicates
from pyadlml.pipeline import Pipeline, FeatureUnion, TrainOnlyWrapper, \
    EvalOnlyWrapper, TrainOrEvalOnlyWrapper, YTransformer
from pyadlml.model_selection import train_test_split
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
import mlflow
from ray.tune.integration.mlflow import MLflowLoggerCallback
from ray import tune
from sklearn.utils import estimator_html_repr
import argparse
from pyadlml.preprocessing.preprocessing import SkTime
from pyadlml.preprocessing.windows import EventWindow

from pyadlml.training.trainable import Trainable 
from ray.tune.integration.mlflow import mlflow_mixin
from ray.air import RunConfig
import ray


tracking_url = "http://localhost:5555"

def main():
    parser=argparse.ArgumentParser(description="train tool")
    parser.add_argument("dataset", type=str)
    args=parser.parse_args()

    seed = 1
    dataset =  args.dataset

    mlflow.set_tracking_uri(tracking_url)
    trainable = Trainable(
        exp_name='masterarbeit',
        ds_name=dataset,
        use_tune=True
    )


    # Define pipeline

    from sktime.classification.hybrid import HIVECOTEV2
    from sktime.classification.dictionary_based import BOSSEnsemble
    from sktime.classification.kernel_based import RocketClassifier
    from sktime.classification.deep_learning import CNNClassifier

    sktime_ev_win_pipe = Pipeline([
        ('enc', StateVectorEncoder()),
        ('lbl', LabelMatcher()),
        ('drop_dups', TrainOnlyWrapper(DropDuplicates())),
        ('drop_time', DropTimeIndex()),
        ('ev_win', EventWindow()),
        ('to_sktime', SkTime(rep='nested_univ')),
        #('clf_rocket', RocketClassifier()),
        ('clf_cnn', CNNClassifier()),
    ])

    trainable.set_pipe(sktime_ev_win_pipe)



    from tools.configs.models import clf_BOSSEnsemble, clf_ROCKET, clf_CNN

    # Define Hyperparameter and parameters for mlflow
    hparam_dict = dict(
        model='CNNClassifier',
        pipe_params=dict(
            enc__encode=tune.choice(['raw', 'changepoint', 'raw+changepoint']),
            lbl__other=True,
            ev_win__rep='many-to-one',
            ev_win__window_size=tune.randint(10, 500),
            ev_win__stride=tune.randint(1,50),
        ),
        mlflow=dict(
            experiment_name=trainable.get_experiment_name(),
            tracking_uri=mlflow.get_tracking_uri(),
        )
    )
    hparam_dict['pipe_params'].update(clf_CNN)


    num_cpus = 4
    num_gpus = 1
    num_samples = 100
    from ray.tune.search.bohb import TuneBOHB
    search_alg = TuneBOHB(metric='val_acc', mode='max')

    ray.init(
        num_cpus=num_cpus,
        num_gpus=num_gpus,
    )
    from ray.air.config import RunConfig
    tuner = tune.Tuner(
        trainable=trainable,
        param_space=hparam_dict,
        tune_config=tune.TuneConfig(
            num_samples=num_samples,
            mode='max',
            metric='val_acc',
            search_alg=search_alg
        ),
        run_config=RunConfig(name="test_mlflow_first")
    )

    analysis = tuner.fit()
    print(analysis)

if __name__ == '__main__':
    main()


from pathlib import Path
from pyadlml.dataset import *
from pyadlml.preprocessing import StateVectorEncoder, LabelMatcher, DropTimeIndex, \
    DropDuplicates
from pyadlml.pipeline import EvalOnlyWrapper, Pipeline, TrainOnlyWrapper
from sklearn.ensemble import RandomForestClassifier
from ray import tune
from pyadlml.training.trainable import Trainable 
import mlflow
import argparse


"""
Train a model 

Dataset avaliable:
    'casas_aruba',
    'amsterdam',
    'kasteren_A',
    'uci_ordonezA',
    'uci_ordonezB',

n    'mitlab_1',
n    'mitlab_2',
n    'kasteren_B',
n    'kasteren_C',
"""


def main():
    parser=argparse.ArgumentParser(description="train tool")
    parser.add_argument("dataset", type=str)
    args=parser.parse_args()

    seed = 1
    dataset =  args.dataset


    mlflow.set_tracking_uri("http://localhost:5555")
    trainable = Trainable(exp_name='masterarbeit', ds_name=dataset)


    # Define Hyperparameter and parameters for mlflow
    hparam_dict = dict(
        model='RandomForestClassifier',
        pipe_params=dict(
            clf__n_estimators=tune.choice([50, 100, 200]),
            clf__warm_start=tune.choice([True, False]),
            clf__max_depth=tune.choice([None, 30, 40]),
            enc__encode='changepoint',
            lbl__other=False
        )
    )

    # Define pipeline
    trainable.set_pipe(
        uci_pipe
    )

    # Sample pipeparams from distribution
    for pname, paramh in hparam_dict['pipe_params'].items():
        try:
            hparam_dict['pipe_params'][pname] = paramh.sample()
        except AttributeError:
            pass

    trainable(hparam_dict)

if __name__ == '__main__':
    main()


import datetime

import numpy as np

PI_CONSTANT = 0.01
TRANS_CONSTANT = 0.01

def gen_rand_transitions(state_count):
    # initalize with random hmm
    trans_matrix = np.random.random_sample((state_count, state_count))
    row_sums = trans_matrix.sum(axis=1)
    trans_matrix = trans_matrix / row_sums[:, np.newaxis]
    return trans_matrix

def gen_rand_emissions(state_count, em_count):
    em_matrix = np.random.random_sample((state_count, em_count))
    row_sums = em_matrix.sum(axis=1)
    em_matrix = em_matrix / row_sums[:, np.newaxis]
    return em_matrix

def gen_rand_pi(state_count):
    init_pi = np.random.random_sample(state_count)
    init_pi = init_pi / sum(init_pi)
    return init_pi

def gen_eq_pi(state_count):
    init_val = 1. / state_count
    return np.full((state_count), init_val)

def gen_handcrafted_priors_for_pi(act_data, K):
    """ calculate the probabilities of beeing in a specific state
    based on the total time spent in the state in the synthetic data
    Parameters
    ----------
    act_data list
        e.g [
                {'name': 6,
                'day_of_week': 0,
                 'start': datetime.time(4, 0),
                 'end': datetime.time(6, 15)},
                {'name': 10, 'day_of_week': 0,
                'start': datetime.time(6, 15),
                'end': datetime.time(8, 45)},
                ....
            ]
    K (int)
        the length of the state vector

    Returns
    -------
    new_pi
        1D K long numpy array
    """
    # total amount of hours spent in this activity
    time_deltas = np.zeros((K), dtype=np.float64)
    total_hours = 0.0
    for act_data_point in act_data:
        # todo inconsistent use of typecast
        #tidx = self._hmm._idx_state(str(act_data_point['name']))
        tidx = act_data_point['name']
        td = _time_diff(act_data_point['end'], act_data_point['start'])
        if act_data_point['end'] == datetime.time.fromisoformat("00:00:00"):
            # if the endtime is 0 o clock the timedifferenz is negativ
            # and has to be corrected
            # example 19:00:00 to 00:00:00 yields -19.0 and corrected is 5.0
            td = td + 24.0
        time_deltas[tidx] += td
        total_hours += td

    new_pi = time_deltas/total_hours # type: np.ndarray
    new_pi = _correct_single_row(new_pi, K)
    return new_pi

def _correct_single_row(new_pi, K):

    """
    assign a constant where the probability is 0.0 and correct the other
    probabilities for it
    """
    cnt_non_zero = np.count_nonzero(new_pi)
    cnt_zero = K - cnt_non_zero
    if cnt_zero > 0:
        non_zero_correction = cnt_zero*PI_CONSTANT/cnt_non_zero
        for i in range(K):
            if new_pi[i] == 0.0:
                new_pi[i] = PI_CONSTANT
            else:
                new_pi[i] -= non_zero_correction
    return new_pi

def _time_diff(a, b):
    """
    computes the differences of two time points
    stackoverflow copy paste
    :param a:
    :param b:
    :return:
        the total amount of time in hours
    """
    dateTimeA = datetime.datetime.combine(datetime.date.today(), a)
    dateTimeB = datetime.datetime.combine(datetime.date.today(), b)

    # Get the difference between datetimes (as timedelta)
    dateTimeDifference = dateTimeA - dateTimeB

    # Divide difference in seconds by number of seconds in hour (3600)
    dateTimeDifferenceInHours = dateTimeDifference.total_seconds() / 3600
    return dateTimeDifferenceInHours


def gen_handcrafted_priors_for_transitions(act_data, K):
    """
    :param K:
        the amount of states
    :return:
        K x K numpy matrix
    """
    sorted_act_data = _sort_act_data(act_data)
    sorted_acts = []
    for act in sorted_act_data:
        sorted_acts.append(act['name'])

    tm = np.zeros((K,K))
    # the total transitions from state i into other states
    norm_trans = np.zeros((K))

    for i in range(1, len(sorted_acts)):
        zn = sorted_acts[i-1]
        znp1 = sorted_acts[i]
        zn_i = zn
        zn_j = znp1
        #zn_i = self._hmm._idx_state(str(zn))
        #zn_j = self._hmm._idx_state(str(znp1))
        tm[zn_i][zn_j] += 1
        norm_trans[zn_i] += 1
    #print(tm)
    #print(norm_trans)
    """
    count the amount of constants used in transition matrix
    this has to be evenly subtracted from the normalized probabilites
    
    """
    #zero_count = K**2 - np.count_nonzero(tm)
    #print(zero_count)
    #constant_correction = zero_count*self._transition_constant/ (K**2 - zero_count)
    #print(constant_correction)
    tm1, norm_tm1 = _stronger_diagonal(tm.copy(), norm_trans)
    tm2 = _correct_transition_matrix(tm1.copy(), norm_tm1)
    return tm2

def _stronger_diagonal(tm, norm_tm):
    """

    Parameters
    ----------
    tm
    norm_tm

    Returns
    -------

    """
    K, _ = tm.shape
    counter = norm_tm.sum()
    scale = int(counter/3)
    # do a third
    norm_tm = norm_tm + scale
    diagonal = np.eye(K, K, dtype=np.float64)
    diagonal = diagonal*scale
    tmres = tm + diagonal

    return tmres, norm_tm

def _correct_transition_matrix(tm, norm_tm):
    K, _ = tm.shape
    for i in range(0, K):
        row_zeros_count = K - np.count_nonzero(tm[i])
        if K == row_zeros_count:
            # if the complete has no activity set it to standard
            tm[i] = np.random.random_sample(K)
            tm[i] = tm[i] / sum(tm[i])
        else:
            row_correction =  row_zeros_count*TRANS_CONSTANT/ (K-row_zeros_count)
            for j in range(0, K):
                if tm[i][j] == 0:
                    # set to nonzero constant
                    tm[i][j] = TRANS_CONSTANT
                else:
                    # normalize with amount of outgoing transitions
                    tm[i][j] = (tm[i][j] / norm_tm[i]) - row_correction
    return tm


def _sort_act_data(act_data):
    """
    sorts the synthetic generated activity data and returns a sorted list
    :return:
    """
    # list of lists, with each sublist representing a day
    # daybins[0] contains all activities for sunday, ...
    day_bins = [[],[],[],[],[],[],[]]
    for syn_act in act_data:
        day_bins[syn_act["day_of_week"]].append(syn_act)
    res_list = []
    for day in day_bins:
        for i in range(len(day)):
            min_idx = _sort_get_min_idx(day)
            res_list.append(day[i])
    return res_list

def _sort_get_min_idx( lst):
    """
    returns the index of the smallest element
    :param lst:
    :return:
    """
    min_idx = 0
    for i in range(1, len(lst)):
        if lst[i]['start'] < lst[min_idx]['start']:
            min_idx = i
    return min_idx



def gen_handcrafted_priors_categorical_emissions(self, K, D):
    """
    :param K:
        the amount of states
    :param D:
        the amount of observations
    :return:
        K x D numpy array
    """
    act_dev_lst = self._gen_pc_em_transform_loc_data()
    #print('loc_data: ', str(self._loc_data))
    #print('dev_lst: ', act_dev_lst)
    #print('-'*10)

    em = np.zeros((K, D))
    #print('em: ', em)
    obs_list = self._hmm._o
    for k, act_dev in enumerate(act_dev_lst):
        #print('#'*100)
        #print('k: ', k)
        #print('act_dev_lst: ', act_dev)
        cnt_ones = len(act_dev)
        #cnt_ones = 0
        #for obs in obs_list:
        #    if obs in act_dev:
        #        cnt_ones += 1
        #print(cnt_ones)
        m = cnt_ones
        if cnt_ones == 0:
            # if there is no observation related to this activity
            # then set the observations for this activity to random values
            em[k] = np.random.random_sample(D)
            em[k] = em[k] / sum(em[k])
        else:
            prob_ones = (1 - (len(obs_list) - m) * self._emission_constant) / m
            #print('prob_ones: ', prob_ones)
            #print('c: ', self._c)
            #print()
            for d, obs in enumerate(obs_list):
                if obs in act_dev:
                    em[k][d] = prob_ones
                else:
                    em[k][d] = self._emission_constant
    return em

def _gen_pc_em_transform_loc_data(self):
    act_state_list = self._hmm._z
    act_dev_lst = []
    #print('state_lst: ', act_state_list)
    #print('dev_lst: ', act_dev_lst)
    #print('loc_data: ', str(self._loc_data))
    """
    [[1,2,3], [2], ... ] 
    with 1,2,3 being the observations for activity 1 
    """
    for idx_z, z in enumerate(act_state_list):
        #print('#'*100)
        #print('act: ', z)
        #print('dev_lst: ', act_dev_lst)
        for loc in self._loc_data:
            #print('*'*10)
            #print('act: ', z)
            #print('lact: ', loc['activities'])
            if z in loc['activities'] and loc['devices'] != []:
                #print('True')
                try:
                    act_dev_lst[idx_z].extend(loc['devices'])
                except:
                    # the first time the there is no list and therefore has
                    # to be assigned
                    #print('idx_z: ', str(idx_z))
                    act_dev_lst.append([])
                    act_dev_lst[idx_z].extend(loc['devices'])
        if len(act_dev_lst)-1 < int(z):
            """
            this is the case if no device was assigned to
            the activity, then an empty list has to be appended
            """
            act_dev_lst.append([])
    return act_dev_lst


from .windows import TimeWindow, EventWindow, ExplicitWindow
from .preprocessing import \
    StateVectorEncoder, \
    IndexEncoder, \
    DropColumn, \
    DropTimeIndex, \
    LabelMatcher, \
    DropDuplicates,\
    Timestamp2Seqtime, \
    KeepOnlyDevices, \
    DropDevices, \
    Identity, \
    SkTime